/*
 *  Simple perfomance counting
 *
 *  v1.0
 *
 *  by Paul.Sokolovsky@technologist.com
 *
 *  This file is in public domain
 *
 */


#ifndef __PERF_COUNT__H
#define __PERF_COUNT__H

#define MHz *1000*1000
#define PC_Freq (100 MHz)

#if defined(BENCHMARK) && defined (__GNUC__) && defined(__i386__)

/* Implementation of perfomance counting for GCC/i586 */

static inline long long get_hw_tsc()
{
  long long a;
  asm volatile("rdtsc":"=A" (a));
  return a;
}

typedef struct
{
  long long start;
  long long ival;
  long count;
} PC_Info;

/* Define perfomance counter */
#if defined(_WIN32) && !defined(DONT_EXPORT_PC) 
#define PC_COUNTER(pc) PC_Info __declspec(dllexport) pc={0,0,0}
#else
#define PC_COUNTER(pc) PC_Info pc={0,0}
#endif

/* Declare external perfomance counter */
#define PC_EXTERN_COUNTER(pc) extern PC_Info pc

/* Declare external perfomance counter residing in other DL-module */
#if defined(_WIN32)
#define PC_DL_COUNTER(pc) extern PC_Info __declspec(dllimport) pc
#else
#define PC_DL_COUNTER(pc) extern PC_Info pc
#endif

/* Initialize perfomance counter */
#define PC_INIT(pc) { pc.ival=pc.start=pc.count=0; }

/* Start(continue) perfomance counting */
#define PC_START(pc) { ++pc.count; pc.start=get_hw_tsc(); }

/* Stop(pause) perfomance counting */
#define PC_STOP(pc) { pc.ival+=get_hw_tsc()-pc.start; }

/* Get value of perfomance counter in ticks */
#define PC_GET(pc) (pc).ival

/* Get value of perfomance counter in seconds */
#define PC_GET_SEC(pc) ((double)(pc).ival/PC_Freq)

/* Get value of perfomance counter in miliseconds */
#define PC_GET_MSEC(pc) ((double)(pc).ival/(PC_Freq/1000))

/* Get times perfomance counter was started */
#define PC_GET_COUNT(pc) (pc).count

/* Print value of perfomance counter in ticks */
#define PC_PRINTF(pc) printf("%f",(double)(pc).ival)

/* Print value of perfomance counter in seconds */
#define PC_PRINTF_SEC(pc) printf("%f",(double)(pc).ival/PC_Freq)

/* Print value of perfomance counter in miliseconds */
#define PC_PRINTF_MSEC(pc) printf("%f",(double)(pc).ival/PC_Freq)

typedef struct _PC_Desc
{
  char *desc;
  PC_Info *pc;
  struct _PC_Desc *children;
} PC_Desc;

#ifdef PC_NO_INCLUDES
/* Live with warnings */
#define PC_FILE void
#else
#include <stdio.h>
#define PC_FILE FILE
#endif


static inline void PC_dump_desc(PC_FILE *f,PC_Desc *desc,int level)
{
  double ival=PC_GET_MSEC(*desc->pc);
  char *fmt="%f msec\n";
  { int i; for (i=level;i--;) fputc(' ',f); }
  if (ival>10000) { ival/=1000; fmt="%f\n"; }
  fprintf(f,"%s\t\t",desc->desc);
  fprintf(f,"%D\t",PC_GET_COUNT(*desc->pc));
  fprintf(f,fmt,ival);
  for (desc=desc->children;desc && desc->desc;desc++)
  				PC_dump_desc(f,desc,level+1);
}

#define PC_REPORT(file,pc_tree) PC_dump_desc(file,pc_tree,0)

#else
/* Either non-benchmark or not supported compiler/platform */
#define PC_COUNTER(pc)
#define PC_EXTERN_COUNTER(pc)
#define PC_DL_COUNTER(pc)
#define PC_INIT(pc)
#define PC_START(pc)
#define PC_STOP(pc)
#define PC_GET(pc)
#define PC_GET_SEC(pc)
#define PC_PRINTF(pc)
#define PC_PRINTF_SEC(pc)
#endif

#endif /* __PERF_COUNT__H */
